#include "test.h"
#include "device_launch_parameters.h"
#include "cuda_runtime.h"
#include <ctime>
#include "chrono"
#include <random>

using namespace std::chrono;

template <typename T>
void printMat(const T &mat, 
              const int &m, 
              const int &n) {
    std::cout << "\n";
    for (int i = 0; i < m; i++)  {
        for (int j = 0; j < n; j++) {
            std::cout << *(mat + i * n + j) << " ";
        }
        std::cout << "\n";
    }
}

template <typename T>
void initMat(const T &mat,
                 const int &m,
                 const int &n) {
    std::default_random_engine engine;
    std::uniform_real_distribution<double> u(0.0, 1.0);
    for (int i = 0; i < m * n; ++i) {
            *(mat + i) = u(engine);
    }
}

template <typename T>
void matMul_cpu(const T& mat_a, 
                const T& mat_b, 
                const T& mat_c,
                const int &m, 
                const int &k,
                const int &n) {
    for (int i = 0; i < m; ++i) {
        for (int j = 0; j < n; ++j) {
            *(mat_c + i * n + j) = 0;
            for (int t = 0; t < k; ++t) {
                *(mat_c + i * n + j) += *(mat_a + i * k + t) * *(mat_b + t * n + j);
            }
        }
    }
}

int main(void) {
    int dev = 0;
    cudaDeviceProp devProp;
    // CHECK(cudaGetDeviceProperties(&devProp, dev));
    cudaGetDeviceProperties(&devProp, dev);
    std::cout << "---------------------------------------------------------" << std::endl;
    std::cout << "GPU device:                           | " << devProp.name << std::endl;
    std::cout << "SM num:                               | " << devProp.multiProcessorCount << std::endl;
    std::cout << "Share Memory Per Block:               | " << devProp.sharedMemPerBlock / 1024 << " KB" << std::endl;
    std::cout << "Max Threads Per Block:                | " << devProp.maxThreadsPerBlock << std::endl;
    std::cout << "Max Threads Per MultiProcessor:       | " << devProp.maxThreadsPerMultiProcessor << std::endl;
    std::cout << "Max Threads Batch Per MultiProcessor: | " << devProp.maxThreadsPerMultiProcessor / 32 << std::endl;
    std::cout << "---------------------------------------------------------" << std::endl;

    int m = 1000;
    int k = 600;
    int n = 1000;
    int nBytes_a = m * k * sizeof(float);
    int nBytes_b = k * n * sizeof(float);
    int nBytes_c = m * n * sizeof(float);

    // use Unified Memory.
    float *mat_a, *mat_b, *mat_c, *mat_r;
    // cudaMallocManaged((void **)&mat_a, nBytes_a);
    // cudaMallocManaged((void **)&mat_b, nBytes_b);
    // cudaMallocManaged((void **)&mat_c, nBytes_c);
    mat_a = (float *)malloc(nBytes_a);
    mat_b = (float *)malloc(nBytes_b);
    mat_c = (float *)malloc(nBytes_c);
    mat_r = (float *)malloc(nBytes_c);
    
    /*
    for (int i = 0; i < m * k; i++) {
        mat_a[i] = 10;
    }
    */
    initMat(mat_a, m, k);
    // printMat(mat_a, m, k);

    /*
    for (int i = 0; i < k * n; i++) {
        mat_b[i] = 10;
    }
    */
    initMat(mat_b, k, n);
    // printMat(mat_b, k, n);

    for (int i = 0; i < m * n; i++) {
        mat_c[i] = 0.0;
    }
    matMul_cpu(mat_a, mat_b, mat_r, m, k, n);
    // printMat(mat_r, m, n);

    float *d_a, *d_b, *d_c;
    cudaMalloc((void **)&d_a, nBytes_a);
    cudaMalloc((void **)&d_b, nBytes_b);
    cudaMalloc((void **)&d_c, nBytes_c);

    auto start = system_clock::now();

    cudaMemcpy((void *)d_a, (void *)mat_a, nBytes_a, cudaMemcpyHostToDevice);
    cudaMemcpy((void *)d_b, (void *)mat_b, nBytes_b, cudaMemcpyHostToDevice);
    cudaMemcpy((void *)d_c, (void *)mat_c, nBytes_c, cudaMemcpyHostToDevice);

    // dim3 blockSize(k);
    dim3 blockSize(1);
    dim3 gridSize(m, n);

    // add<<<gridSize, blockSize>>>(x, y, z, N);
    MatMul<<<gridSize, blockSize>>>(d_a, d_b, d_c, m, k, n);
    cudaError_t cudaStatus = cudaGetLastError();
    if (cudaStatus != cudaSuccess) {
        std::cout << "cudaError: " << cudaGetErrorString(cudaStatus) << std::endl;
    }

    cudaMemcpy((void *)mat_c, (void *)d_c, nBytes_c, cudaMemcpyDeviceToHost);
    cudaDeviceSynchronize();

    auto end = system_clock::now();
    auto duration = duration_cast<microseconds>(end - start);
    
    std::cout << "GPU time: "
         << double(duration.count()) * microseconds::period::num / microseconds::period::den << " seconds" << std::endl;

    // printMat(mat_c, m, n);
    float maxError = 0.0;
    for (int i = 0; i < m * n; i++) {
        maxError = fmax(maxError, fabs(*(mat_c + i) - *(mat_r + i)));
    }
    std::cout << "Max Error: " << maxError << std::endl;

    cudaFree(d_a);
    cudaFree(d_b);
    cudaFree(d_c);

    free(mat_a);
    free(mat_b);
    free(mat_c);

    // test<<<1, 1>>>();
    // std::cout << "Hello CUDA!" << std::endl;
    return 0;
}
